## A large-scale stochastic spatiotemporal model for Aedes albopictus-borne chikungunya epidemiology
## -----------------------------------------------------------------------------------------------------------
## 
## S5 Text: Posterior sampling
## ===========================
##
## Here, we aim to describe the use of the hoppMCMC and albopictus python packages to simulate and perform parameter inferences for the chikungunya epidemiological model using the provided environmental and surveillance datasets.
##
## We begin by importing required libraries (see https://pypi.python.org/pypi for missing packages)
import numpy
import json
import hoppMCMC
import albopictus as aa
##
## We, then, import environmental variables and the surveillance dataset
data = json.load(open("dataItaly.json","r"))
##
## data is a python dictionary which contains the following information:
##     clim: Array of Python dictionaries containing "dates" and "envar" for vector dynamics simulations
##           dates: Sequence of numbers (days) indicating the duration of a simulation
##           envar: Concatenation of photoperiod, daily mean air temperature, daily precipitation, and human population density covering the desired time frame of the simulation
##           In the case of Italy, clim contains only one environmental dataset associated with a single patch (see the main text). The environmental dataset covers 826 days from 16 June 2006 to 18 September 2008
##     tprobs: Matrix of transfer probabilities between patches. Diagonal elements represent the daily probability of staying in a patch. Non diagonal entries will be normalised per row to yield a row sum of 1
##           In the case of Italy, since there is only one patch, the probability of residing in it is 1
##     t0, tf: Array indices marking the beginning and end of chikungunya transmission simulation
##     obs: Array of reported infectious cases
##     std: Array of error terms affiliated with the observations
##     posmat: Matrix to be multiplied with simulation results to yield an array of incidences with dates corresponding to the dates of the observations
##     par: Python dictionary with "par0", a sample within the posterior mode of interest, and "varmat", the transition kernel for the MCMC algorithm for posterior sampling. Initial parameter configuration and transition kernel are first guesses from the optimisation results of the hoppMCMC algorithm (see below)
##
## We simulate vector abundance using the posterior parameter samples from posterior mode Q4
sims = [aa.vector.simPar(data['clim'],pr) for pr in aa.param['Q4']]
## Then, we construct the environmental variables dictionary using the simulation results. For improving efficiency, we only consider the dates from 31 May 2007 to 18 August 2007
climchikv = [{
    'dates': data['clim']['dates'][data['t0']:data['tf']],
    'envar': numpy.hstack([
        numpy.mean([s['coln4f'][data['t0']:data['tf']] for s in sims],axis=0), # daily average adult female
        numpy.mean([s['cold4'][data['t0']:data['tf']] for s in sims],axis=0), # average adult life expectancy
        numpy.mean([s['cold4s'][data['t0']:data['tf']] for s in sims],axis=0), # standard deviation of adult life expectancy
        numpy.mean([s['colF4'][data['t0']:data['tf']] for s in sims],axis=0), # daily average fecundity
        4000]) # total number of people (all susceptible)
    }]
##
## We define the domain of parameter search with this function
def checkParam(pr):
    if (pr[aa.chikv.parids['introduce_time']]<1 # Time of introduction of an infectious person (with regard to the beginning of simulation)
        or pr[aa.chikv.parids['introduce_time']]>365
        or pr[aa.chikv.parids['gamma']]<0 # Infectious period in humans
        or pr[aa.chikv.parids['gamma']]>10
        or pr[aa.chikv.parids['omega']]<0 # Latent period in humans
        or pr[aa.chikv.parids['omega']]>10
        or pr[aa.chikv.parids['vec_omega']]<0 # Latent period in vectors
        or pr[aa.chikv.parids['vec_omega']]>10
        or pr[aa.chikv.parids['gamma_sd']]<0 # Standard deviation of gamma
        or pr[aa.chikv.parids['gamma_sd']]>5
        or pr[aa.chikv.parids['omega_sd']]<0 # Standard deviation of omega
        or pr[aa.chikv.parids['omega_sd']]>5
        or pr[aa.chikv.parids['vec_omega_sd']]<0 # Standard deviation of vec_omega
        or pr[aa.chikv.parids['vec_omega_sd']]>5
        or pr[aa.chikv.parids['susceptH']]<0 # Human susceptibility to infections
        or pr[aa.chikv.parids['susceptH']]>1
        or pr[aa.chikv.parids['susceptV']]<0 # Vector susceptibility to infections
        or pr[aa.chikv.parids['susceptV']]>1
        or pr[aa.chikv.parids['prob_chronic']]<0 # Probability of a persistent infection following the initial infectious period
        or pr[aa.chikv.parids['prob_chronic']]>1
        or pr[aa.chikv.parids['prob_C_to_Ic']]<0 # Daily probability of relapse during a chronic infection
        or pr[aa.chikv.parids['prob_C_to_Ic']]>365
        or pr[aa.chikv.parids['prob_C_to_R']]<0 # Daily probability of recovery from a chronic infection
        or pr[aa.chikv.parids['prob_C_to_R']]>365
        or (pr[aa.chikv.parids['prob_C_to_Ic']]+pr[aa.chikv.parids['prob_C_to_R']])>365
        or pr[aa.chikv.parids['ovitraps']]<0 # Average number of ovitraps per square kilometer
        or pr[aa.chikv.parids['ovitraps']]>1e3
        or pr[aa.chikv.parids['report']]<0 # Fraction of symptomatic infections which are also reported
        or pr[aa.chikv.parids['report']]>1
        or pr[aa.chikv.parids['qmove']]<0 # Probability of a person to reside in a patch for a day
        or pr[aa.chikv.parids['qmove']]>1
        or pr[aa.chikv.parids['connect']]<0 # Degree of separation among patches
        or pr[aa.chikv.parids['connect']]>5
        or pr[aa.chikv.parids['Khum']]<0 # Average number of bites per human
        or pr[aa.chikv.parids['Khum']]>100
        or pr[aa.chikv.parids['Kvec']]<0 # Average number of bites per vector and daily fecundity
        or pr[aa.chikv.parids['Kvec']]>1):
        return True
    return False
##
## We define the Chi-square distance function with the error term sigma. The term T is included in the hoppMCMC algorithm as the annealing temperature 
def fitnessChikv(pr):
    # First, we check if we reside in the search domain
    if checkParam(pr):
        return numpy.Inf
    #
    # Then, we simulate a trajectory
    allsims = aa.chikv.simSpread(climchikv,data['tprobs'],pr)
    # We calculate the total number of reported and recorded incidents in all patches
    sim = numpy.sum([s['colnInew'] for s in allsims],axis=0)
    # We match dates to the time frame of the incidence data (days, weeks, months, etc.)
    m = sim.dot(data['posmat'])
    # We avoid cases of no outbreak with a high distance value
    if all(m<=1.0):
        return 1e13
    # Finally, we return the Chi-square distance between the data and simulations
    return 0.5*sum(((data['obs'] - m)/data['std'])**2)
##
## The score function can be calculated for any parameter configuration. We test it with "par0" below
print fitnessChikv(data['par']['par0'])
##
## In order to perform inference, we define an array with the indices of parameters to be inferred
inferpar = numpy.array([aa.chikv.parids[x] for x in [
    'gamma',
    'gamma_sd',
    'omega',
    'omega_sd',
    'vec_omega',
    'vec_omega_sd',
    'susceptH',
    'susceptV',
    'prob_chronic',
    'prob_C_to_Ic',
    'prob_C_to_R',
    'ovitraps',
    'report',
    'Khum',
    'Kvec'
]])
##
## We execute the adaptive basin-hopping algorithm with the following configuration. We note that the initial condition is arbitrary, and its effect on finding the optimum posterior mode is negligible
hoppMCMC.hoppMCMC(fitnessChikv,
                  [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 4.5, 1.0, 3.5, 0.5, 0.7, 0.33, 1.0, 1.0, 0.8, 3.0, 0.1, 640.0, 14.0, 0.0, 1.0, 0.0, 0.0, 0.5, 0.5, 3.2, 50.0, 0.25],
                  numpy.diag(numpy.repeat(1e-14,len(inferpar))),
                  inferpar=inferpar,
                  gibbs=True,
                  rangeT=[1,1e3],
                  model_comp=1e3,
                  num_hopp=10,
                  num_adapt=50,
                  num_chain=12,
                  chain_length=50,
                  outfilename=None)
##
## Once we obtain a parameter configuration from a desired posterior mode (data['par']['par0']), we run individual MCMC chains beginning with this configuration. Here, we defined an arbitrary transition kernel (data['par']['varmat']) by iterating data['par']['par0'] a few steps and calculating the emerging covariance of the Markov chain. With the following configuration, we exploit an adaptive transition kernel (scaled every 100 iterations) to yield an acceptance rate of about 1%
mcmc = hoppMCMC.chainMCMC(fitnessChikv,
                    data['par']['par0'],
                    data['par']['varmat'],
                    gibbs=True,
                    pulsevar=1.0,
                    anneal=2.0,
                    accthr=0.0,
                    inferpar=inferpar,
                    varmat_change=0,
                    pulse_change=100,
                    pulse_change_ratio=2,
                    print_iter=0)
##
## We run the chain for 20000 iterations. Here, we print out each iteration, but this information should be stored in the file system for later access
for m in range(20000):
    print m
    mcmc.iterate()
    print mcmc.getParam()
##
## In order to sample from a posterior mode, we iterated 50 chains in parallel, discarded the initial 10000 iterations, and extracted 20 from each with regular intervals
##
